/*****************************************************************
  Filename:  ascat_l1b_reader.c

  This is a library for reading ASCAT L1B files in "native format" aka PFS aka EPS

  Modified 12-16 Feb 2010 by DGL at BYU to read newer file formats and SZF files
  Modified 30 Apr 2010 by DGL at BYU to include orbit end
  Modified 7 July 2011 by RDL at BYU: Small cleanups to reduce gcc warnings
   (unused variables, unused function, comparison with types of different
    signedness, no return type specified for a few functions,
    no return value for a non-void function)
  Updated by RDL at BYU 15 May 2013:
      version 2.0: Lots of code cleanups and changes, updated for v12 L1B files
  Updated by RDL at BYU 29 May 2013:
      Bugfixes, convert C string to Fortran for the relevant Fortran functions
  Updated by RDL at BYU 26 Mar 2014:
      Return time in millisecond resolution instead of seconds
  Updated by RDL at BYU 4 Apr 2014:
      Return next/previous nodes in along-track direction as well
  Updated by RDL at BYU 6 May 2014:
      Along-track node positions fixed for v12 format data
  Updated by RDL at BYU 14 Aug 2014:
      Flip ascending/descending flag to match the L1B data, not the L1B documentation
  Updated by RDL at BYU 27 Oct 2014:
      Add along-beam node index (different numbering for v11 and v12 format data)
  Updated by RDL at BYU 04 Nov 2014:
      Corrected along-beam index for side beams
  Updated by RDL at BYU 23 Dec 2014:
      Check quality flags for v12 format data

******************************************************************/

/********************************************************************/
// Documentation
//
// This is compiled as a library (static or shared) and called to read (no
// writing) ASCAT L1B files. Both C and Fortran interfaces are provided.
//
// It provides the public functions:
//
// ascat_check_file():   basic check that the L1B file is ok
// ascat_open_file():    open a file for reading
// ascat_next_mdr():     move to next mdr (measurement data record) in the file
// ascat_next_node():    return node data for the next node in the current MDR
// ascat_get_sat():      return satellite id
// ascat_get_fmt():      return format id
// ascat_get_tot_mdr():  return number of MDRs in the file
// ascat_get_tot_node(): return number of nodes per beam
// ascat_clear_node():   initialize all of the node members to zero
// ascat_close_file():   close the file and do any cleanup needed
//
// Some private functions exist behind the scenes.
// Also some structs are used to contain all of the relevant data together.
//
// The basic structure of an L1B file is as follows:
// There is a Main Product Header Record (MPHR) that contains information about the rest of the file
// There are other headers that we mostly ignore
// Then there is a bunch of Measurement Data Records (MDR)
// For L1B format version 11 and below, each MDR contains the measurements of a single sequence of the six beams
// For L1B format version 12 and above, each MDR contains the measurements for only a single beam, so there are six times as many MDRs now
// This distinction is abstracted behind the ascat_next_node() function, just keep calling it until it returns false (0)
//
/********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <math.h>

#include "ascat_l1b_reader.h"

/********************************************************************/
// Useful definitions

// record id

#define MPHR_ID   1
#define SPHR_ID   2
#define IPR_ID    3
#define GEADR_ID  4
#define GIADR_ID  5
#define VEADR_ID  6
#define VIADR_ID  7
#define MDR_ID    8
#define ASCAT_ID  2
#define SZO_ID    2
#define SZR_ID    1
#define SZF_ID    3
#define DUMMY_ID 13

// record size in bytes

#define GRH_SIZE        20
#define MPHR_SIZE     3307
#define IPR_SIZE        27
#define SZO_SIZE      4018
#define SZR_SIZE      7818
#define SZF_SIZE     41624 // for v11 format and below
#define SZF_SIZE_NEW  3684 // for v12 format and up

// number of nodes

#define NODES_SZO       21
#define NODES_SZR       41
#define NODES_SZF      256 // for v11 format and below
#define NODES_SZF_NEW  192 // for v12 format and up

// throwpos function

#define THROW return throw( __FILE__, __FUNCTION__, __LINE__ )

// Throw a message
static int throw(const char *file, const char *func, int line) {
  fprintf(stderr, " *** ASCAT ERROR in file %s, %s at line %i\n", file, func, line);
  return 1;
}

// Helpful macros
#define FREE(p)   do { free(p); (p) = NULL; } while(0)

// Function prototypes that are private to this file
static int get_next_szf_node(const ascat_file_t * const a, ascat_node_t * const b);

/********************************************************************/
// Fortran interface
//
// NB: Fortran passes by reference, so the fortran interface makes this
// adjustment
//
// Since it's too difficult to represent the types ascat_file, ascat_node, and ascat_orbit in Fortran, I hide it by using global static variables.

static ascat_file_t *f_a_file;
static ascat_node_t *f_a_node;
static ascat_orbit_t *f_a_orbit;

int ascat_check_file_(char *fname, size_t fname_len) {
  char *end;

  // Trim trailing whitespace
  // (Convert Fortran string to C string)
  end=fname + fname_len - 1;
  while (end > fname && isspace(*end)) end--;
  *(end+1) = 0;

  return ascat_check_file(fname);
}

int ascat_open_file_(char *fname, size_t fname_len) {
  char *end;

  // Trim trailing whitespace
  // (Convert Fortran string to C string)
  end=fname + fname_len - 1;
  while (end > fname && isspace(*end)) end--;
  *(end+1) = 0;

  // Allocate the static variables for use
  f_a_file = malloc(sizeof(ascat_file_t));
  if (f_a_file == NULL) {
    fprintf(stderr, "*** ERROR: ascat_open_file_: cannot allocate %zd bytes for f_a_file\n", sizeof(ascat_file_t));
    THROW;
  }
  f_a_node = malloc(sizeof(ascat_node_t));
  if (f_a_node == NULL) {
    fprintf(stderr, "*** ERROR: ascat_open_file_: cannot allocate %zd bytes for f_a_node\n", sizeof(ascat_node_t));
    THROW;
  }
  f_a_orbit = malloc(sizeof(ascat_orbit_t));
  if (f_a_orbit == NULL) {
    fprintf(stderr, "*** ERROR: ascat_open_file_: cannot allocate %zd bytes for f_a_orbit\n", sizeof(ascat_orbit_t));
    THROW;
  }

  return ascat_open_file(fname, f_a_file, f_a_orbit);
}

void ascat_close_file_() {
  ascat_close_file(f_a_file);

  FREE(f_a_orbit);
  FREE(f_a_node);
  FREE(f_a_file);
  return;
}

int ascat_get_tot_mdr_() {
  return (int) ascat_get_tot_mdr(f_a_file);
}

int ascat_get_tot_node_() {
  return (int) ascat_get_tot_node(f_a_file);
}

int ascat_get_node_beam_() {
  return (int) ascat_get_node_beam(f_a_file);
}

void ascat_get_fmt_(char *fmt, int fmt_len) {
  // Note that this returns a Fortran string

  char *fmt_tmp;
  fmt_tmp = ascat_get_fmt(f_a_file);
  strncpy(fmt, fmt_tmp, fmt_len);

  // Replace the null characters with spaces
  memset(fmt+strlen(fmt), ' ', fmt_len-strlen(fmt));

  return;
}

void ascat_get_sat_(char *sat, int sat_len) {
  // Note that this returns a Fortran string

  switch (ascat_get_sat(f_a_file)) {
    case METOP1:
      strncpy(sat, "MetOp-B", sat_len);
      break;
    case METOP2:
      strncpy(sat, "MetOp-A", sat_len);
      break;
    case METOP3:
      strncpy(sat, "MetOp-C", sat_len);
      break;
    default:
      strncpy(sat, "Unknown", sat_len);
      break;
  }

  // Replace the null characters with spaces
  memset(sat+strlen(sat), ' ', sat_len-strlen(sat));

  return;
}

int ascat_next_mdr_() {
  return ascat_next_mdr(f_a_file, f_a_node);
}

int ascat_next_node_(ascat_node_t *b) {
  int ret_val;

  ret_val = ascat_next_node(f_a_file, f_a_node);

  b->sigma0 = f_a_node->sigma0;
  b->inc = f_a_node->inc;
  b->azi = f_a_node->azi;
  b->lat = f_a_node->lat;
  b->lon = f_a_node->lon;
  b->sat = f_a_node->sat;
  b->fmt = f_a_node->fmt;
  b->asc = f_a_node->asc;
  b->tm = f_a_node->tm;
  b->swath = f_a_node->swath;
  b->index = f_a_node->index;
  b->ab_index = f_a_node->index;
  b->beam_prev_lat = f_a_node->beam_prev_lat;
  b->beam_prev_lon = f_a_node->beam_prev_lon;
  b->beam_next_lat = f_a_node->beam_next_lat;
  b->beam_next_lon = f_a_node->beam_next_lon;
  b->track_prev_lat = f_a_node->track_prev_lat;
  b->track_prev_lon = f_a_node->track_prev_lon;
  b->track_next_lat = f_a_node->track_next_lat;
  b->track_next_lon = f_a_node->track_next_lon;

  return ret_val;
}

/********************************************************************/
// Low-level functions to extract info from the MDR

static void get_bit(const unsigned char *x, int pos, unsigned char *a, unsigned int bit) {
	// x: pointer to data
	// pos: offset where byte is
	// a: bit return value (it's either 1 or 0)
  // bit: which bit number to access (1-8)
  // 	1 (0000 0001)
  // 	2 (0000 0010)
  // 	4 (0000 0100)
  // 	8 (0000 1000)
  // 16 (0001 0000)
  // 32 (0010 0000)
  // 64 (0100 0000)

  unsigned char byte, bitmask;

  byte = *(x+pos);
  bitmask = 1 << (bit-1);
  *a = (byte & bitmask) != 0;
}

static void get_char(const unsigned char *x, int pos, int *a) {
  *a = *(x+pos);
  if( *a >= 128 ) *a -= 256;
}

// Decodes 2-byte unsigned integers
static void get_ushort(const unsigned char *x, int pos, int *a) {
  x += pos;
  *a = *x;
  x++;
  *a = *a*256 + (*x);
}

static void get_short(const unsigned char *x, int pos, int *a) {
  x += pos;
  *a = *x;
  if( *a >= 128 ) *a -= 256;
  x++;
  *a = *a*256 + (*x);
}

static int get_uint(const unsigned char *x, int pos, int *a) {
  // Decodes a 4-byte unsigned integer
  //
  // x: the pointer to a location within the buffer
  // pos: offset where the uint starts relative to x
  // a: return the uint4
  // x is not modified by this function (though it appears to at first glance;
  // however, it is a pointer passed by value, not a uchar passed by reference)
  // Returns 0 for successful decode
  //
  x += pos;
  *a = *x;
  if( *a > 128 ) THROW;
  x++;
  *a = *a*256 + (*x);
  x++;
  *a = *a*256 + (*x);
  x++;
  *a = *a*256 + (*x);
  return 0;
}

// This function gets called quite a bit
// Decodes 4-byte integers
static inline void get_int(const unsigned char *x, int pos, int *a) {
  x += pos;
  *a = *x;
  if( *a > 128 ) *a -= 256;
  x++;
  *a = *a*256 + (*x);
  x++;
  *a = *a*256 + (*x);
  x++;
  *a = *a*256 + (*x);
}

// If I do use this function, remove the attribute-ed prototype
static void get_uchar3(const unsigned char *x, int pos, int *a, int *b, int *c) __attribute__ ((unused));
static void get_uchar3(const unsigned char *x, int pos, int *a, int *b, int *c) {
  x += pos;
  *a = *x;
  x++;
  *b = *x;
  x++;
  *c = *x;
}

static int get_time(const unsigned char *x, double *t) {
  // Long CDS time
  // Used for pre-version 12 data
  int day,msec;

  get_ushort(x,20,&day);
  if( get_uint(x,22,&msec) ) return 1;
  // microseconds is stored too, but we don't care about microsecond resolution
  // get_ushort(x,26,&usec);
  //sec = msec/1000;
  *t = day + msec/(1000 * 86400.0);
  return 0;
}

static int get_cds_time(const unsigned char *x, double *t) {
  // Short CDS time
  // Used for version 12 data
  int day,msec;

  // Note the offset is different in the MDR than the v11 data
  get_ushort(x,22,&day);
  if( get_uint(x,24,&msec) ) return 1;
  //sec = msec/1000;
  //printf("day: %d, sec: %d, msec: %d\n", day, sec, msec);
  //*t = day + sec/86400.0;
  *t = day + msec/(1000 * 86400.0);
  return 0;
}

static void get_ushort1(const unsigned char *x, int pos, double sf, double *ans) {
  int a;

  get_ushort(x,pos,&a);
  *ans = a*sf;
}

static void get_short1(const unsigned char *x, int pos, double sf, double *ans) {
  int a;

  get_short(x,pos,&a);
  *ans = a*sf;
}

static void get_int1(const unsigned char *x, int pos, double sf, double *ans) {
  int a;

  get_int(x,pos,&a);
  *ans = a*sf;
}

// If I do use this function, remove the attribute-ed prototype
static void get_ushort3(const unsigned char *x, int pos, double sf, double *a, double *b, double *c) __attribute__ ((unused));
static void get_ushort3(const unsigned char *x, int pos, double sf, double *a, double *b, double *c) {
  int ans;

  get_ushort(x,pos,&ans);
  *a = ans*sf;
  get_ushort(x,pos+2,&ans);
  *b = ans*sf;
  get_ushort(x,pos+4,&ans);
  *c = ans*sf;
}

// If I do use this function, remove the attribute-ed prototype
static void get_short3(const unsigned char *x, int pos, double sf, double *a, double *b, double *c) __attribute__ ((unused));
static void get_short3(const unsigned char *x, int pos, double sf, double *a, double *b, double *c) {
  int ans;

  get_short(x,pos,&ans);
  *a = ans*sf;
  get_short(x,pos+2,&ans);
  *b = ans*sf;
  get_short(x,pos+4,&ans);
  *c = ans*sf;
}

// If I do use this function, remove the attribute-ed prototype
static void get_int3(const unsigned char *x, int pos, double sf, double *a, double *b, double *c) __attribute__ ((unused));
static void get_int3(const unsigned char *x, int pos, double sf, double *a, double *b, double *c) {
  int ans;

  get_int(x,pos,&ans);
  *a = ans*sf;
  get_int(x,pos+4,&ans);
  *b = ans*sf;
  get_int(x,pos+8,&ans);
  *c = ans*sf;
}

static int get_line(const unsigned char *x, int n, char *a) {
  int cnt=1;
  while( n ) {
    *a = (char)( *x );
    if (*a == 10) {  // end of line encountered
      *a = '\0';
      n=1;
    } else
      cnt++;
    x++;
    a++;
    n--;
  }
  *a = '\0';
  return(cnt);
}

static void get_str(const unsigned char *x, int n, char *a) {
  while( n ) {
    *a = (char)( *x );
    x++;
    a++;
    n--;
  }
  *a = '\0';
}

static int get_string(const unsigned char *x, int pos, int n, char *ans) {
  if( n > 18 ) THROW;
  if( n < 0 ) THROW;
  x += pos;
  get_str(x,n,ans);
  return 0;
}

static int get_number(const unsigned char *x, int pos, int *ans) {
  char s[16];

  x += pos;
  get_str(x,6,s);
  if( sscanf(s,"%i",ans) != 1 ) THROW;
  if( *ans < 0 ) THROW;
  return 0;
}

static int get_double(const char *x, double *ans) {
  if( sscanf(x,"%lf",ans) != 1 ) THROW;
  return 0;
}


static int get_field_string(const unsigned char *x, const char *fname, char *fstring, int fslen) {
  int n=0, l=0, m;
  char *p;
  char line[256];

  while( n < MPHR_SIZE) {
    m=get_line(x+n,256,line);
    n += m;
    l++;
    //printf("line %d (%d) %s\n",l,m,line);
    if( (p=strstr(line,fname)) != 0)
      if( (p=strchr(p,'=')) != 0) {
        if( get_string((unsigned char *)p, 2, fslen, fstring) ) THROW;
        //printf(" %d extracted value: '%s'\n",l,fstring);
        return 0;
      }
  }
  return 1;
}

static int get_field_value(const unsigned char *x, const char *fname, char *fstring, int fslen, double *fvalue, float scale) {
  int n=0, l=0, m;
  char *p;
  char line[256];

  while( n < MPHR_SIZE) {
    m=get_line(x+n,256,line);
    n += m;
    l++;
    //printf("line %d (%d) %s\n",l,m,line);
    if( (p=strstr(line,fname)) != 0)
      if( (p=strchr(p,'=')) != 0) {
        if( get_string((unsigned char *)p, 2, fslen, fstring) ) THROW;
        *fvalue=0;
        if( get_double(fstring,fvalue) ) THROW;
        *fvalue = *fvalue / (double) scale;
        //printf(" %d extracted value: %s (%lf)\n",l,fstring,*fvalue);
        return 0;
      }
  }
  return 1;
}

// record handling
static int read_rec(FILE *fp, int size, char id, unsigned char *x ) {
  int n;

  if( fread(x,1,size,fp) != (unsigned int) size ) THROW;
  if( *x != id ) THROW;
  get_uint(x,4,&n);
  if( n != size ) THROW;
  return 0;
}

static int skip_rec(FILE *fp, char id) {
  unsigned char x[GRH_SIZE];
  int n;

  if( fread(x,1,GRH_SIZE,fp) != GRH_SIZE ) THROW;
  if( *x != id ) THROW;
  get_uint(x,4,&n);
  if( n < GRH_SIZE ) THROW;
  n -= GRH_SIZE;
  if( fseek(fp,n,SEEK_CUR) ) THROW;
  return 0;
}

/********************************************************************/
// High-level functions

#define A_FAIL(msg) { fprintf(stderr, " *** ascat_check_file, line %d: %s\n", __LINE__, msg); fclose(fp); return 1; }
int ascat_check_file(const char *fname) {
  // basic check that the L1B file is ok
  // Returns 0 if the file looks good, 1 otherwise

  char instr[16],prod[16];
  unsigned char x[MPHR_SIZE];
  int num_mphr,num_sphr,skip;
  int num_geadr,num_giadr;
  int num_veadr,num_viadr;
  int num_ipr,num_mdr;
  int i,n;
  int fmt_ver;
  int mdr_size,id;
  FILE *fp;

  fp = fopen(fname,"rb");
  if( fp == NULL ) {
    fprintf(stderr, " *** ERROR: can't open %s for reading\n", fname);
    perror("fopen error");
    return 1;
  }

  // Check MPHR
  if( read_rec(fp,MPHR_SIZE,MPHR_ID,x) )
    A_FAIL("bad mphr, are you sure this is a PFS-format file?");
  if( get_string(x,552,4,instr) )
    A_FAIL("bad instr string");
  if( get_string(x,625,3,prod) )
    A_FAIL("bad prod string");
  if( get_number(x,1036,&fmt_ver) )
    A_FAIL("bad fmt ver");
  if( get_number(x,2714,&num_mphr) )
    A_FAIL("bad num_mphr");
  if( get_number(x,2753,&num_sphr) )
    A_FAIL("bad num_sphr");
  if( get_number(x,2792,&num_ipr) )
    A_FAIL("bad num_ipr");
  if( get_number(x,2831,&num_geadr) )
    A_FAIL("bad num_geadr");
  if( get_number(x,2870,&num_giadr) )
    A_FAIL("bad num_giadr");
  if( get_number(x,2909,&num_veadr) )
    A_FAIL("bad num_veadr");
  if( get_number(x,2948,&num_viadr) )
    A_FAIL("bad num_viadr");
  if( get_number(x,2987,&num_mdr) )
    A_FAIL("bad num_mdr");

  // Check for right number of headers
  if( num_mphr != 1 )
    A_FAIL("bad num_mphr value");
  if (fmt_ver >= 11) {
    // v11 has 1 SPHR, older has 0 SPHR
    if( num_sphr != 1 )
      A_FAIL("bad num_sphr value");
  }
  else {
    if( num_sphr != 0 )
      A_FAIL("bad num_sphr value");
  }

  if( strcmp(instr,"ASCA") )
    A_FAIL("bad instr value");

  if( !strcmp(prod,"SZO") ) {
    mdr_size = SZO_SIZE;
    id = SZO_ID;
  } else if( !strcmp(prod,"SZR") ) {
    mdr_size = SZR_SIZE;
    id = SZR_ID;
  } else if( !strcmp(prod,"SZF") ) {
    if (fmt_ver >= 12)
      mdr_size = SZF_SIZE_NEW;
    else
      mdr_size = SZF_SIZE;
    id = SZF_ID;
  } else
    A_FAIL("bad prod value");

  // skip other headers
  // sphr (if format version 11+)
  if( fmt_ver >= 11 )
    for (i=0; i<num_sphr; i++)
      if( skip_rec(fp,SPHR_ID) )
        A_FAIL("bad sphr");

  // ipr
  for(i=0;i<num_ipr;i++)
    if( skip_rec(fp,IPR_ID) )
      A_FAIL("bad ipr");

  // geadr
  for(i=0;i<num_geadr;i++)
    if( skip_rec(fp,GEADR_ID) )
      A_FAIL("bad geadr");

  // giadr
  for(i=0;i<num_giadr;i++)
    if( skip_rec(fp,GIADR_ID) )
      A_FAIL("bad giadr");

  // veadr
  for(i=0;i<num_veadr;i++)
    if( skip_rec(fp,VEADR_ID) )
      A_FAIL("bad veadr");

  // viadr
  for(i=0;i<num_viadr;i++)
    if( skip_rec(fp,VIADR_ID) )
      A_FAIL("bad viadr");

  // check all mdr's
  skip = mdr_size - GRH_SIZE;
  for(i=0;i<num_mdr;i++) {
    if( fread(x,1,GRH_SIZE,fp) != GRH_SIZE )
      A_FAIL("mdr fread");

    if( *x != MDR_ID )
      A_FAIL("mdr id");

    // Sometimes EUMETSAT puts MDRs with the ID set to
    // Dummy (13).  These records indicate gaps in time.
    // We need to skip over these.
    if ( *(x+1) == DUMMY_ID) {
      // Find size of dummy MDR
      get_uint(x, 4, &n);
      n -= GRH_SIZE;
      // Skip it
      if( fseek(fp,n,SEEK_CUR) )
        A_FAIL("dummy mdr fseek");
      fprintf(stdout, "*** INFO: dummy MDR detected (MDR %d)\n", i); // debug
      continue;
    }
    if( *( x + 1 ) != ASCAT_ID )
      A_FAIL("mdr not ascat");

    if( *( x + 2 ) != id )
      A_FAIL("mdr - unexpected subclass (SZO/SZR/SZF)");

    get_uint(x,4,&n);
    if( n != mdr_size )
      A_FAIL("mdr - unexpected record size");

    // Go to next MDR
    if( fseek(fp,skip,SEEK_CUR) )
      A_FAIL("mdr fseek");
  }

  fclose(fp);
  return 0;
}

int ascat_open_file(const char *fname, ascat_file_t *a, ascat_orbit_t *orb) {
  // open a file for reading and extract some info from the headers

  unsigned char x[MPHR_SIZE];
  char instr[5],prod[4],satid[4];
  int num_mphr,num_sphr;
  int num_geadr,num_giadr;
  int num_veadr,num_viadr;
  int num_ipr,num_mdr;
  int fmt_ver;
  int pver_maj, pver_min;
  int i, n;
  char tstring[19];
  double tval;

  // open file
  a->fp = fopen(fname,"rb");
  if( a->fp == NULL ) {
    fprintf(stderr, " *** ERROR: can't open %s for reading\n", fname);
    perror("fopen error");
    THROW;
  }

  // mphr
  if( read_rec(a->fp,MPHR_SIZE,MPHR_ID,x) ) THROW;

  // extract variables of interest by name in the mphr record
  if( get_field_string(x,"INSTRUMENT_ID", instr, 4) ) THROW;
  if( get_field_string(x,"PRODUCT_TYPE", prod, 3) ) THROW;
  if( get_field_string(x,"SPACECRAFT_ID", satid, 3) ) THROW;
  if( get_field_string(x,"SENSING_START", orb->time_start, 15) ) THROW;
  if( get_field_string(x,"SENSING_END", orb->time_stop, 15) ) THROW;
  if( get_field_value(x,"ORBIT_START", tstring, 16, &tval, 1.0) ) THROW;
  orb->orbit_start=(int) tval;
  if( get_field_value(x,"ORBIT_END", tstring, 16, &tval, 1.0) ) THROW;
  orb->orbit_end=(int) tval;
  if( get_field_string(x,"STATE_VECTOR_TIME",orb->state_time, 18) ) THROW;
  if( get_field_value(x,"SEMI_MAJOR_AXIS", tstring, 18, &orb->semi, 1.e3) ) THROW;
  if( get_field_value(x,"ECCENTRICITY", tstring, 18, &orb->eccen, 1.e6) ) THROW;
  if( get_field_value(x,"INCLINATION", tstring, 18, &orb->inclination, 1.e3) ) THROW;
  if( get_field_value(x,"PERIGEE_ARGUMENT", tstring, 18, &orb->perigee, 1.e3) ) THROW;
  if( get_field_value(x,"RIGHT_ASCENSION", tstring, 18, &orb->ascension, 1.e3) ) THROW;
  if( get_field_value(x,"MEAN_ANOMALY", tstring, 18, &orb->mean_anomaly, 1.e3) ) THROW;
  if( get_field_value(x,"X_POSITION", tstring, 18, &orb->x_position, 1.e3) ) THROW;
  if( get_field_value(x,"Y_POSITION", tstring, 18, &orb->y_position, 1.e3) ) THROW;
  if( get_field_value(x,"Z_POSITION", tstring, 18, &orb->z_position, 1.e3) ) THROW;
  if( get_field_value(x,"X_VELOCITY", tstring, 18, &orb->x_vel, 1.e3) ) THROW;
  if( get_field_value(x,"Y_VELOCITY", tstring, 18, &orb->y_vel, 1.e3) ) THROW;
  if( get_field_value(x,"Z_VELOCITY", tstring, 18, &orb->z_vel, 1.e3) ) THROW;
  if( get_field_value(x,"SUBSAT_LATITUDE_START", tstring, 18, &orb->startlat, 1.e3) ) THROW;
  if( get_field_value(x,"SUBSAT_LONGITUDE_START", tstring, 18, &orb->startlon, 1.e3) ) THROW;
  if( get_field_value(x,"SUBSAT_LATITUDE_END", tstring, 18, &orb->stoplat, 1.e3) ) THROW;
  if( get_field_value(x,"SUBSAT_LONGITUDE_END", tstring, 18, &orb->stoplon, 1.e3) ) THROW;


  //printf("Instrument string=%s\n",instr);
  //printf("Product string=%s\n",prod);
  //printf("SatID string=%s\n",satid);

  /*
     printf("Orbit start=%d\n",orb->orbit_start);
     printf("Start Time=%s\n",orb->time_start);
     printf("Stop Time=%s\n",orb->time_stop);
     printf("State Time=%s\n",orb->state_time);
     printf("Semi Major Axis=%lf\n",orb->semi);
     printf("Eccentricity=%lf\n",orb->eccen);
     printf("Inclination Angle=%lf\n",orb->inclination);
     printf("Argument of Perigee=%lf\n",orb->perigee);
     printf("Right Ascension Angle=%lf\n",orb->ascension);
     printf("Mean Anomaly=%lf\n",orb->mean_anomaly);
     printf("X position=%lf\n",orb->x_position);
     printf("Y position=%lf\n",orb->y_position);
     printf("Z position=%lf\n",orb->z_position);
     */

  //if( get_string(x,552,4,instr) ) THROW; // old way
  //if( get_string(x,625,3,prod) ) THROW;
  //if( get_string(x,696,3,satid) ) THROW;
  //printf("Instr string=%s\n",instr);
  //printf("Prod string=%s\n",prod);
  //printf("Satid string=%s\n",satid);

  // extract variables from the mphr record
  if( get_number(x,1036,&fmt_ver) ) THROW;
  if( get_number(x,2714,&num_mphr) ) THROW;
  if( get_number(x,2753,&num_sphr) ) THROW;
  if( get_number(x,2792,&num_ipr) ) THROW;
  if( get_number(x,2831,&num_geadr) ) THROW;
  if( get_number(x,2870,&num_giadr) ) THROW;
  if( get_number(x,2909,&num_veadr) ) THROW;
  if( get_number(x,2948,&num_viadr) ) THROW;
  if( get_number(x,2987,&num_mdr) ) THROW;

  if( get_number(x,962,&pver_maj) ) THROW;
  if( get_number(x,1000,&pver_min) ) THROW;
  //if( get_field_value(x,"PROCESSOR_MAJOR_VERSION", &pver_maj, 5) ) THROW;
  //if( get_field_value(x,"PROCESSOR_MINOR_VERSION", &pver_min, 5) ) THROW;
  a->pver = pver_maj * 100 + pver_min;
  a->fver = fmt_ver;

  // check values
  if( num_mphr != 1 ) THROW;
  if( fmt_ver >= 11 ) {
    // new ASCAT L1B version has 1 for sphr
    if( num_sphr != 1 ) THROW;
  } else {
    if( num_sphr != 0 ) THROW;
  }

  if( strcmp(instr,"ASCA") ) THROW;

  if( !strcmp(prod,"SZO") ) {
    a->fmt = SZO;
    a->fmt_id = SZO_ID;
    a->nodes_per_beam = NODES_SZO;
    a->mdr_size = SZO_SIZE;
  } else if( !strcmp(prod,"SZR") ) {
    a->fmt = SZR;
    a->fmt_id = SZR_ID;
    a->nodes_per_beam = NODES_SZR;
    a->mdr_size = SZR_SIZE;
  } else if( !strcmp(prod,"SZF") ) {
    a->fmt = SZF;
    a->fmt_id = SZF_ID;
    if (fmt_ver >= 12) {
      a->nodes_per_beam = NODES_SZF_NEW;
      a->mdr_size = SZF_SIZE_NEW;
    } else {
      a->nodes_per_beam = NODES_SZF;
      a->mdr_size = SZF_SIZE;
    }
  } else
    THROW;

  // skip header records
  if( fmt_ver >= 11 )
    for (i=0; i<num_sphr; i++)
      if( skip_rec(a->fp,SPHR_ID) ) THROW;

  for(i=0;i<num_ipr;i++)
    if( skip_rec(a->fp,IPR_ID) ) THROW;

  for(i=0;i<num_geadr;i++)
    if( skip_rec(a->fp,GEADR_ID) ) THROW;

  for(i=0;i<num_giadr;i++)
    if( skip_rec(a->fp,GIADR_ID) ) THROW;

  for(i=0;i<num_veadr;i++)
    if( skip_rec(a->fp,VEADR_ID) ) THROW;

  for(i=0;i<num_viadr;i++)
    if( skip_rec(a->fp,VIADR_ID) ) THROW;

  // misc
  if( !strcmp(satid,"M01") )
    a->sat = METOP1;
  else if( !strcmp(satid,"M02") )
    a->sat = METOP2;
  else if( !strcmp(satid,"M03") )
    a->sat = METOP3;
  else
    a->sat = METOPN;

  // Allocate space to store the rest of the L1B file
  a->num_mdrs = num_mdr;
  a->mdrs = malloc( num_mdr * a->mdr_size * sizeof(unsigned char));
  if (a->mdrs == NULL) {
    fprintf(stderr, "*** ERROR: ascat_open_file: cannot allocate %zd bytes for a->mdrs\n", num_mdr * a->mdr_size * sizeof(unsigned char));
    THROW;
  }
  a->mdr = a->mdrs;

  // Read each MDR into the allocated space, skip dummy MDRs
  // (we're done with the x buffer for the MPHR, so we re-use it for reading the GRH for each MDR)
  for(i=0;i<num_mdr;i++) {
    // Read in the general record header
    if( fread(x, 1, GRH_SIZE, a->fp) != GRH_SIZE ) THROW;

    // First byte of GRH: what class it is. We check to see if it's an MDR
    if( x[0] != MDR_ID ) THROW;

    // Sometimes EUMETSAT puts MDRs with the ID set to
    // Dummy (13).  These records indicate gaps in time.
    // We need to skip over these.
    // Second byte of GRH: what instrument
    if ( x[1] == DUMMY_ID) {
      // Find size of dummy MDR
      get_uint(x, 4, &n);
      n -= GRH_SIZE;
      // Skip the dummy MDR
      if( fseek(a->fp, n, SEEK_CUR) ) {
        perror("fseek error");
        THROW;
      }
      a->num_mdrs--;
      fprintf(stdout, "*** INFO: dummy MDR detected (MDR %d)\n", i); // debug
      continue;
    }
    if( x[1] != ASCAT_ID ) THROW;

    // Third byte of GRH: record subclass (SZO/SZR/SZF for ASCAT)
    if( x[2] != a->fmt_id ) THROW;

    // Bytes 5-8 of GRH: the record size (as an uint4)
    get_uint(x, 4, &n);
    if( n != (int) a->mdr_size ) THROW;

    // Check MDR quality flags (this is not in the GRH already read in)
    // We check the status of these flags but we don't do anything about them other than warn
    if (a->fver >= 12) {
      char flag;
      // DEGRADED_INST_MDR
      fread(&flag, sizeof(char), 1, a->fp);
      if (flag == 1)
        printf("*** WARNING: MDR %d has degraded quality (instrument)\n", i);

      //DEGRADED_PROC_MDR
      fread(&flag, sizeof(char), 1, a->fp);
      if (flag == 1)
        printf("*** WARNING: MDR %d has degraded quality (processing)\n", i);

      // Rewind the file pointer by two bytes (the two quality flags)
      fseek(a->fp, -2 * sizeof(char), SEEK_CUR);
    }

    // Okay, we have a good MDR, so let's copy it (including its GRH) into our mdrs buffer
    memcpy(a->mdr, x, GRH_SIZE);
    if (fread(a->mdr+GRH_SIZE, 1, a->mdr_size-GRH_SIZE, a->fp) != a->mdr_size-GRH_SIZE) THROW;
    a->mdr += a->mdr_size;
  }

  // We've loaded all the MDRs into memory, reset mdr pointer to NULL for the
  // first call to ascat_next_mdr()
  a->mdr = NULL;
  a->cur_mdr = 0;

  return 0;
}

int ascat_next_mdr(ascat_file_t * const a, ascat_node_t * const b) {
  // move to next mdr (measurement data record) in the file
  //
  // Updates the pointer in the ascat_file_t struct
  // Returns:
  //    0: success
  //    1: failure
  //    2: we've reached EOF, no more MDRs
  //
  // Note that in order to "rewind" to the first MDR, all that is needed is to
  // set a->mdr to NULL and then call this function, since that acts like the
  // first call.
  //

  // Note that cur_mdr goes from 0 to num_mdrs-1
  if (a->mdr == NULL) {
    // This is the first call to next_mdr
    a->cur_mdr = 0;
    a->mdr = a->mdrs;

    ascat_clear_node(b);
    return 0;
  } else if (a->cur_mdr + 1 >= a->num_mdrs) {
    // No more MDRs
    return 2;
  } else if (a->mdr - a->mdrs + a->mdr_size > a->num_mdrs * a->mdr_size)
    // We'll go beyond the buffer
    return 2;
  else {
    // All other MDRs
    a->mdr += a->mdr_size;
    a->cur_mdr++;

    ascat_clear_node(b);
    return 0;
  }
}

int ascat_next_node(const ascat_file_t * const a, ascat_node_t * const node) {
  // return node data for the next node in the current MDR
  // Returns:
  //    0: success
  //    1: failure
  //    2: end of MDR, no more nodes

  //clear_ascat_node(b); // This is helpful for debugging, but takes a bit too much time in real life
  switch( a->fmt ) {
    case SZO:
      //if( get_szo_node(a->mdr,i,swath,node) ) return 1;
      //b->asc = calc_asc( b->track );
      fprintf(stderr, "*** Error: SZO files not implemented currently\n");
      THROW;
      break;
    case SZR:
      //if( get_szr_node(a->mdr,i,swath,node) ) return 1;
      //b->asc = calc_asc( b->track );
      fprintf(stderr, "*** Error: SZR files not implemented currently\n");
      THROW;
      break;
    case SZF:
      return get_next_szf_node(a,node);
      break;
    default:
      THROW;
  }
  return 1;
}

static int get_next_szf_node(const ascat_file_t * const a, ascat_node_t * const b) {
  // Get the next ASCAT SZF node in the current MDR
  // Returns:
  //    0: success
  //    1: failure
  //    2: end of MDR, no more nodes

  unsigned int node_i, beam_i;
  int offset;
  int ans_int;
  double ans_f;

  // Set some node variables
  b->sat = a->sat;
  b->fmt = a->fmt;

  // Find the current beam, node index
  node_i = b->index;
  beam_i = b->swath;

  // check indices are ok
  //if( node_i < 0 ) THROW; // node_i is unsigned
  if( node_i > a->nodes_per_beam ) THROW; // 256 for v11, 192 for v12
  if (beam_i == 0) {
    //fprintf(stdout, " starting mdr from beginning\n");
    beam_i = 1;
  }
  if( beam_i < 1 ) THROW;
  if( beam_i > 6 ) THROW;

  // Increment to next node
  if (a->fver < 12) {
    // All six beams are in the same MDR
    node_i++;
    if (node_i > a->nodes_per_beam) {
      node_i = 1;
      beam_i++;
      if (beam_i > 6) {
        b->swath = 1;
        b->index = 0;
        return 2;
      }
    }
  } else {
    // Only a single beam is in an MDR
    node_i++;
    // b->index and b->swath are initialized when we start a new MDR, so they
    // won't hold the previous values; instead we read swath number from the MDR
    get_char(a->mdr, 31, (int *) &beam_i);
    if (node_i > a->nodes_per_beam) {
      //b->swath++;
      //b->index = 0;
      //if (b->swath > 6) b->swath = 1;
      return 2;
    }
  }
  b->index = node_i;
  b->swath = beam_i;
  // The v12 format has the first and last several nodes removed from each beam. We use the "ab_index" variable to have a consistent numbering.
  if (a->fver < 12)
    b->ab_index = b->index;
  else {
    if (b->swath == 2 || b->swath == 5) {
      // According to Craig Anderson at EUMETSAT, for the mid beams, only
      // samples 32-223 (0-based indexing) are written. Note that we're using
      // 1-based indexing, so if we have an index of 1, the along-beam index
      // (sample) is 33.
      b->ab_index = b->index + 32;
    }
    else if (b->swath == 1 || b->swath == 4 || b->swath == 3 || b->swath == 6) {
      // According to Craig Anderson at EUMETSAT, for the side beams, only
      // samples 38-229 (0-based indexing) are written. Note that we're using
      // 1-based indexing, so if we have an index of 1, the along-beam index
      // (sample) is 38.
      b->ab_index = b->index + 38;
    } else {
      fprintf(stderr, "swath %d is invalid\n", b->swath);
      THROW;
    }
  }

  // Quality flags
  // We only check these for the v12 format L1B data
  if (a->fver >= 12) {
    unsigned char f_noise, f_pgp, f_filter, v_filter;
    unsigned char f_pg, v_pg, f_pgp_drop, f_np;
    unsigned char f_orbit, f_attitude, f_omega, f_man, f_osv, f_e_tel_pres, f_tel_ir;
    unsigned char f_sol, f_geo;

    int quality;

    // These flags occur once per MDR, so I could speed things up by caching
    // them or something. Oh well, this probably isn't that time consuming
    // anyway.

    // FLAGFIELD_RF1
    get_bit(a->mdr, 3488, &f_noise, 1);
    get_bit(a->mdr, 3488, &f_pg, 2);
    get_bit(a->mdr, 3488, &v_pg, 3);
    get_bit(a->mdr, 3488, &f_filter, 4);
    get_bit(a->mdr, 3488, &v_filter, 5);

    // FLAGFIELD_RF2
    get_bit(a->mdr, 3489, &f_pgp, 1);
    get_bit(a->mdr, 3489, &f_np, 2);
    get_bit(a->mdr, 3489, &f_pgp_drop, 3);

    // FLAGFIELD_PL
    get_bit(a->mdr, 3490, &f_orbit, 1);
    get_bit(a->mdr, 3490, &f_attitude, 2);
    get_bit(a->mdr, 3490, &f_omega, 3);
    get_bit(a->mdr, 3490, &f_man, 4);
    get_bit(a->mdr, 3490, &f_osv, 5);

    // FLAGFIELD_GEN1
    get_bit(a->mdr, 3491, &f_e_tel_pres, 1);
    get_bit(a->mdr, 3491, &f_tel_ir, 2);

    // f_ce, v_ce, f_oa, f_tel, f_ref are also in flagfield_gen1 and serve as
    // an aggregation of some of the previous flags

    // These flags occur for every measurement node
    // FLAGFIELD_GEN2
    get_bit(a->mdr, 3492 + (node_i-1), &f_sol, 1);
    get_bit(a->mdr, 3492 + (node_i-1), &f_geo, 3);

    // Okay, we've read in all the flags. According to Craig Anderson at
    // EUMETSAT, the flags may be categorized into "red" and "amber" to
    // indicate severity. Amber flags indicate that the data is slightly
    // degraded but still usable. Red flags indicate that the data is severely
    // degraded and should be discarded or used with caution.
    // Red flags: v_pg, v_filter, f_np, f_orbit, f_attitude, f_omega, f_man, f_tel_ir, f_geo, f_pgp
    // Amber flags: f_noise, f_pg, f_filter, f_sol
    // These are warnings since they may mean other flags are incorrect: f_osv, f_e_tel_pres
    // He didn't mention this one but I interpret it as warning: f_pgp_drop
    //
    // Summarize the flags into: bad, warn, good
    if (v_pg || v_filter || f_np || f_orbit || f_attitude || f_omega || f_man || f_tel_ir || f_geo || f_pgp)
      quality = 2;
    else if (f_noise || f_pg || f_filter || f_sol || f_osv || f_e_tel_pres || f_pgp_drop)
      quality = 1;
    else
      quality = 0;

//    if (quality == 1) {
//      // Display warning but continue
//      fprintf(stderr, "WARNING: degraded quality measurement (%d of MDR %zd); flags set: ", node_i, a->cur_mdr);
//      if (f_noise)
//        fprintf(stderr, "f_noise ");
//      if (f_pg)
//        fprintf(stderr, "f_pg ");
//      if (f_filter)
//        fprintf(stderr, "f_filter ");
//      if (f_sol)
//        fprintf(stderr, "f_sol ");
//      if (f_osv)
//        fprintf(stderr, "f_osv ");
//      if (f_e_tel_pres)
//        fprintf(stderr, "f_e_tel_pres ");
//      if (f_pgp_drop)
//        fprintf(stderr, "f_pgp_drop ");
//      fprintf(stderr, "\n");
//
//      // fprintf(stderr, "WARNING: degraded quality measurement (%d of MDR %zd); f_noise: %d f_pg: %d f_filter: %d f_sol: %d f_osv: %d f_e_tel_pres: %d f_pgp_drop: %d\n", node_i, a->cur_mdr, f_noise, f_pg, f_filter, f_sol, f_osv, f_e_tel_pres, f_pgp_drop);
//
//    } else if (quality == 2) {
//      // Display warning and skip measurement
//      fprintf(stderr, "WARNING: skipping bad quality measurement (%d of MDR %zd); flags set: ", node_i, a->cur_mdr);
//      if (v_pg)
//        fprintf(stderr, "v_pg ");
//      if (v_filter)
//        fprintf(stderr, "v_filter ");
//      if (f_np)
//        fprintf(stderr, "f_np ");
//      if (f_orbit)
//        fprintf(stderr, "f_orbit ");
//      if (f_attitude)
//        fprintf(stderr, "f_attitude ");
//      if (f_omega)
//        fprintf(stderr, "f_omega ");
//      if (f_man)
//        fprintf(stderr, "f_man ");
//      if (f_tel_ir)
//        fprintf(stderr, "f_tel_ir ");
//      if (f_geo)
//        fprintf(stderr, "f_geo ");
//      if (f_pgp)
//        fprintf(stderr, "f_pgp ");
//      fprintf(stderr, "\n");
//
//      // fprintf(stderr, "WARNING: skipping bad quality measurement (%d of MDR %zd); v_pg: %d v_filter: %d f_np: %d f_orbit: %d f_attitude: %d f_omega: %d f_man: %d f_tel_ir: %d f_geo: %d f_pgp: %d\n", node_i, a->cur_mdr, v_pg, v_filter, f_np, f_orbit, f_attitude, f_omega, f_man, f_tel_ir, f_geo, f_pgp);
//
//      // Recurse to the next node
//      return get_next_szf_node(a, b);
//    }

    // Skip measurement is quality is bad
    if (quality == 2) {
      // Recurse to the next node
      return get_next_szf_node(a, b);
    }
  } // end if fver

  // ascending/descending flag
  // Note that in the L1B format documenation, a "0" means ascending and a "1"
  // means descending. However, this appears to be reversed. So a 1 means
  // ascending and a 0 means descending. This is the same meaning as I use the "asc" member.
  if (a->fver < 12) {
    get_char(a->mdr, 116+(beam_i-1), &ans_int);
    if( ans_int == 1 )
      b->asc = 1; // True
    else
      b->asc = 0; // False
  } else {
    get_char(a->mdr, 30, &ans_int);
    if( ans_int == 1 )
      b->asc = 1; // True
    else
      b->asc = 0; // False
  }

  // extract node data from mdr
  if (a->fver < 12) {
    if( get_time(a->mdr + 8*(beam_i-1), &b->tm ) ) return 1;
    //printf("time=%f\n",b->tm);

    // find the index to the node in the mdr
    offset = (beam_i-1)*a->nodes_per_beam + node_i - 1;
    //printf("c-- node_i: %d, beam_i: %d, offset: %d\n", node_i, beam_i, offset);

    get_int1(a->mdr, 18560 + offset*4, 1.0e-6, &b->lat);
    get_int1(a->mdr, 24704 + offset*4, 1.0e-6, &b->lon);
    get_int1(a->mdr, 128 + offset*4, 1.0e-6, &b->sigma0);
    get_int1(a->mdr, 6272 + offset*4, 1.0e-6, &b->inc);
    get_int1(a->mdr, 12416 + offset*4, 1.0e-6, &b->azi);
    // The land flag is untested, but should in principle work
    //get_bit(x, 40088 + i*4, &b->fland0, 7); // land flag
  } else {
    if (get_cds_time(a->mdr, &b->tm)) return 1;
    //printf("time=%f\n",b->tm);

    // find the index to the node in the mdr
    offset = node_i - 1;

    // Read in the values
    get_int1(a->mdr, 1568 + offset*4, 1.0e-6, &b->lat);
    get_int1(a->mdr, 2336 + offset*4, 1.0e-6, &b->lon);
    get_int1(a->mdr, 32 + offset*4, 1.0e-6, &b->sigma0);
    get_ushort1(a->mdr, 800 + offset*2, 1.0e-2, &b->inc);
    get_short1(a->mdr, 1184 + offset*2, 1.0e-2, &b->azi);
    // The land fraction is not accurate, so we don't use it
    get_ushort1(a->mdr, 3104 + offset*2, 1.0e-2, &ans_f);
  }

  // Old SRF estimates required the next/prev lat/lon values in both the
  // along-beam and along-track direction. That is no longer required but the
  // code here still exists just in case.
  //
  // For the along-beam values, we can just use the next/previous measurement node in the MDR.
  // For the along-track values, we use the next/previous MDR. For v11 and
  // below, we can do the next/previous MDR since each MDR holds all six beams.
  // For v12 and above, each MDR only holds one beam. So to get to the next
  // track measurement for the same beam we have to skip to SIX MDRs ahead.

  // along-beam next lat,lon
  if( node_i < a->nodes_per_beam ) {
    if (a->fver < 12) {
      get_int1(a->mdr, 18560 + (offset+1)*4, 1.0e-6, &b->beam_next_lat);
      get_int1(a->mdr, 24704 + (offset+1)*4, 1.0e-6, &b->beam_next_lon);
    } else {
      get_int1(a->mdr, 1568 + (offset+1)*4, 1.0e-6, &b->beam_next_lat);
      get_int1(a->mdr, 2336 + (offset+1)*4, 1.0e-6, &b->beam_next_lon);
    }
  } else {
    // at last node, so duplicate data
    b->beam_next_lat = b->lat;
    b->beam_next_lon = b->lon;
  }

  // along-beam prev lat,lon
  if( node_i > 1 ) {
    if (a->fver < 12) {
      get_int1(a->mdr, 18560 + (offset-1)*4, 1.0e-6, &b->beam_prev_lat);
      get_int1(a->mdr, 24704 + (offset-1)*4, 1.0e-6, &b->beam_prev_lon);
    } else {
      get_int1(a->mdr, 1568 + (offset-1)*4, 1.0e-6, &b->beam_prev_lat);
      get_int1(a->mdr, 2336 + (offset-1)*4, 1.0e-6, &b->beam_prev_lon);
    }
  } else {
    // at first node, so duplicate data
    b->beam_prev_lat = b->lat;
    b->beam_prev_lon = b->lon;
  }

  // along-track next lat,lon
  if (a->fver < 12) {
    if( a->cur_mdr+1 < a->num_mdrs ) {
      // Next MDR
      get_int1(a->mdr + a->mdr_size, 18560 + offset*4, 1.0e-6, &b->track_next_lat);
      get_int1(a->mdr + a->mdr_size, 24704 + offset*4, 1.0e-6, &b->track_next_lon);
    } else {
      // at last MDR, so duplicate data
      b->track_next_lat = b->lat;
      b->track_next_lon = b->lon;
    }
  } else {
    if( a->cur_mdr+6 < a->num_mdrs ) {
      get_int1(a->mdr + 6*a->mdr_size, 1568 + offset*4, 1.0e-6, &b->track_next_lat);
      get_int1(a->mdr + 6*a->mdr_size, 2336 + offset*4, 1.0e-6, &b->track_next_lon);
    } else {
      // at last group of six MDRs, so duplicate data
      b->track_next_lat = b->lat;
      b->track_next_lon = b->lon;
    }
  }

  // along-track prev lat,lon
  if (a->fver < 12) {
    if( a->cur_mdr > 0 ) {
      // Previous MDR
      get_int1(a->mdr - a->mdr_size, 18560 + offset*4, 1.0e-6, &b->track_prev_lat);
      get_int1(a->mdr - a->mdr_size, 24704 + offset*4, 1.0e-6, &b->track_prev_lon);
    } else {
      // at first MDR, so duplicate data
      b->track_prev_lat = b->lat;
      b->track_prev_lon = b->lon;
    }
  } else {
    if( a->cur_mdr > 6 ) {
      get_int1(a->mdr - 6*a->mdr_size, 1568 + offset*4, 1.0e-6, &b->track_prev_lat);
      get_int1(a->mdr - 6*a->mdr_size, 2336 + offset*4, 1.0e-6, &b->track_prev_lon);
    } else {
      // at first group of six MDRs, so duplicate data
      b->track_prev_lat = b->lat;
      b->track_prev_lon = b->lon;
    }
  }

  //fprintf(stdout, "c-- swath: %d, node: %d, lat: %f, lon: %f, sigma0: %f, inc: %f, azi: %f, asc: %d, land_frac: %f\n", b->swath, b->index, b->lat, b->lon, b->sigma0, b->inc, b->azi, b->asc, ans_f);
  //fprintf(stdout, "c-- prev lat/lon: %f, %f, next lat/lon: %f, %f\n", b->prev_lat, b->prev_lon, b->next_lat, b->next_lon);
 return 0;
}

unsigned int ascat_get_sat(const ascat_file_t *a) {
  // return satellite id
  return a->sat;
}

char *ascat_get_fmt(const ascat_file_t *a) {
  // return format id
  switch (a->fmt) {
    case(SZO):
      return "SZO";
      break;
    case (SZR):
      return "SZR";
      break;
    case (SZF):
      return "SZF";
      break;
    default:
      return "UNKNOWN";
      break;
  }
}

size_t ascat_get_tot_mdr(const ascat_file_t *a) {
  // return number of MDRs
  return a->num_mdrs;
}

size_t ascat_get_node_beam(const ascat_file_t *a) {
  // return total number of nodes per beam
  return a->nodes_per_beam;
}

size_t ascat_get_node_mdr(const ascat_file_t *a) {
  // return number of nodes per MDR
  if (a->fver >= 12)
    return a->nodes_per_beam;
  else
    switch (a->fmt) {
      case SZO:
      case SZR:
        return a->nodes_per_beam * 2;
        break;
      case SZF:
        return a->nodes_per_beam * 6;
        break;
      default:
        return 0;
    }
}

size_t ascat_get_tot_node(const ascat_file_t *a) {
  // return total number of nodes in the file
  return ascat_get_node_mdr(a) * a->num_mdrs;
}

void ascat_clear_node(ascat_node_t *node) {
  // initialize all of the node members to zero
  node->sat = node->fmt = node->asc = node->swath = node->index = node->ab_index = 0;
  node->lon = node->lat = node->tm = node->track = 0.0;
  node->sigma0 = node->inc = node->azi = 0.0;
  return;
}

void ascat_close_file(ascat_file_t *a) {
  // close the file and do any cleanup needed
  fclose(a->fp);
  FREE(a->mdrs);
  return;
}

/////////////////////////////////////
//// calc ascending flag from track
/////////////////////////////////////
//
//static int calc_asc( double track )
//{
// if( track < 0.0 ) track += 360.0;
// if( track>90.0 && track<270.0 )
//  return 0;
// return 1;
//}
//
/////////////////////////////////////
//// get node data
/////////////////////////////////////
//
//static int get_szo_node( unsigned char *x, int i, int swath, ascat_node *b )
//{
// // check node index is ok
//
// if( i < 0 ) THROW;
// if( i >= NODES_SZO ) THROW;
//
// // use the node index and
// // swath value to calculate the
// // index of the data in the mdr
//
// if( swath )
//  i += NODES_SZO;
// else
//  i = NODES_SZO - 1 - i;
//
// // extract node data from mdr
//
// if( get_time( x, &b->tm ) ) return 1;
//
// get_ushort1( x, 26, 1.0e-2, &b->track );
// get_int1( x, 154 + i*4, 1.0e-6, &b->lat );
// get_int1( x, 322 + i*4, 1.0e-6, &b->lon );
// get_int3( x, 742 + i*12, 1.0e-6, &b->s0, &b->s1, &b->s2 );
// get_ushort3( x, 1246 + i*6, 1.0e-4, &b->kp0, &b->kp1, &b->kp2 );
// get_ushort3( x, 1498 + i*6, 1.0e-2, &b->t0, &b->t1, &b->t2 );
// get_short3( x, 1750 + i*6, 1.0e-2, &b->a0, &b->a1, &b->a2 );
// get_uchar3( x, 2002 + i*3, &b->fkp0, &b->fkp1, &b->fkp2 );
// get_uchar3( x, 2128 + i*3, &b->fuse0, &b->fuse1, &b->fuse2 );
// get_ushort3( x, 2254 + i*6, 1.0e-3, &b->fsyn0, &b->fsyn1, &b->fsyn2 );
// get_ushort3( x, 2506 + i*6, 1.0e-3, &b->fsynq0, &b->fsynq1, &b->fsynq2 );
// get_ushort3( x, 2758 + i*6, 1.0e-3, &b->forb0, &b->forb1, &b->forb2 );
// get_ushort3( x, 3010 + i*6, 1.0e-3, &b->fsol0, &b->fsol1, &b->fsol2 );
// get_ushort3( x, 3262 + i*6, 1.0e-3, &b->ftel0, &b->ftel1, &b->ftel2 );
// get_ushort3( x, 3514 + i*6, 1.0e-3, &b->fext0, &b->fext1, &b->fext2 );
// get_ushort3( x, 3766 + i*6, 1.0e-3, &b->fland0, &b->fland1, &b->fland2 );
// return 0;
//}
//
//static int get_szr_node( unsigned char *x, int i, int swath, ascat_node *b )
//{
// // check node index is ok
//
// if( i < 0 ) THROW;
// if( i >= NODES_SZR ) THROW;
//
// // use the node index and
// // swath value to calculate the
// // index of the data in the mdr
//
// if( swath )
//  i += NODES_SZR;
// else
//  i = NODES_SZR - 1 - i;
//
// // extract node data from mdr
//
// if( get_time( x, &b->tm ) ) return 1;
// get_ushort1( x, 26, 1.0e-2, &b->track );
// get_int1( x, 274 + i*4, 1.0e-6, &b->lat );
// get_int1( x, 602 + i*4, 1.0e-6, &b->lon );
// get_int3( x, 1422 + i*12, 1.0e-6, &b->s0, &b->s1, &b->s2 );
// get_ushort3( x, 2406 + i*6, 1.0e-4, &b->kp0, &b->kp1, &b->kp2 );
// get_ushort3( x, 2898 + i*6, 1.0e-2, &b->t0, &b->t1, &b->t2 );
// get_short3( x, 3390 + i*6, 1.0e-2, &b->a0, &b->a1, &b->a2 );
// get_uchar3( x, 3882 + i*3, &b->fkp0, &b->fkp1, &b->fkp2 );
// get_uchar3( x, 4128 + i*3, &b->fuse0, &b->fuse1, &b->fuse2 );
// get_ushort3( x, 4374 + i*6, 1.0e-3, &b->fsyn0, &b->fsyn1, &b->fsyn2 );
// get_ushort3( x, 4866 + i*6, 1.0e-3, &b->fsynq0, &b->fsynq1, &b->fsynq2 );
// get_ushort3( x, 5358 + i*6, 1.0e-3, &b->forb0, &b->forb1, &b->forb2 );
// get_ushort3( x, 5850 + i*6, 1.0e-3, &b->fsol0, &b->fsol1, &b->fsol2 );
// get_ushort3( x, 6342 + i*6, 1.0e-3, &b->ftel0, &b->ftel1, &b->ftel2 );
// get_ushort3( x, 6834 + i*6, 1.0e-3, &b->fext0, &b->fext1, &b->fext2 );
// get_ushort3( x, 7326 + i*6, 1.0e-3, &b->fland0, &b->fland1, &b->fland2 );
// return 0;
//}
//
//
//static int get_szf_node( unsigned char *x, int index, int swath, ascat_node *b )
//{
// int i;
//
// // check node index is ok
// if( index < 0 ) THROW;
// if( index >= NODES_SZF ) THROW;
//
// // use the node index and
// // swath value to calculate the
// // index of the data in the mdr
// i = swath*NODES_SZF + index - 1;
//
// // asc/dsc flag
// get_ushort1( x, 116+swath, 1.0e0, &b->a0 );
// if( b->a0 == 1 )
//   b->asc = 0; // False
// else
//   b->asc = 1; // True
//
// // extract node data from mdr
// if( get_time( x+8*swath, &b->tm ) ) return 1;
// //printf("time=%f\n",b->tm);
//
// //get_ushort1( x, 26, 1.0e-2, &b->track );
// get_int1( x, 18560 + i*4, 1.0e-6, &b->lat );
// get_int1( x, 24704 + i*4, 1.0e-6, &b->lon );
// get_int1( x, 128 + i*4, 1.0e-6, &b->s0 ); // sigma0
// //get_ushort3( x, 2406 + i*6, 1.0e-4, &b->kp0, &b->kp1, &b->kp2 );
// get_int1( x, 6272 + i*4, 1.0e-6, &b->t0 ); // incidence angle
// get_int1( x, 12416 + i*4, 1.0e-6, &b->a0 ); // azimuth angle
// //get_bit( x, 40088 + i*4, &b->fland0, 7); /* land flag */
// // The land flag is untested, but should in principle work
//
// // since we have to compute slice shape (and we have the storage locations available)
// // we need the next/prev lat/lon values, so recover and save these, too.
// if( index < NODES_SZF ) { // next lat,lon
//   i = swath*NODES_SZF + index - 1 + 1;
//   get_int1( x, 18560 + i*4, 1.0e-6, &b->a2 );
//   get_int1( x, 24704 + i*4, 1.0e-6, &b->t2 );
// } else { // at last node, so duplicate data
//   b->a2 = b->lat;
//   b->t2 = b->lon;
// }
// if( index >1 ) {         // prev lat,lon
//   i = swath*NODES_SZF + index - 1 - 1;
//   get_int1( x, 18560 + i*4, 1.0e-6, &b->a1 );
//   get_int1( x, 24704 + i*4, 1.0e-6, &b->t1 );
// } else { // at first node, so duplicate data
//   b->a1 = b->lat;
//   b->t1 = b->lon;
// }
//
// return 0;
//}
//
//
// vim:et:tabstop=2:softtabstop=2:shiftwidth=2:
