/*****************************************************************
  Filename:  ascat_l1b_reader.h

  This is a library for reading ASCAT L1B files
******************************************************************/
#pragma once
#include <stdio.h>

/********************************************************************/
// Documentation
//
// This is compiled as a library (static or shared) and called to read (no
// writing) ASCAT L1B files. Both C and Fortran interfaces are provided.
//
// It provides the public functions:
//
// ascat_check_file():    basic check that the L1B file is ok
// ascat_open_file():     open a file for reading
// ascat_next_mdr():      move to next mdr (measurement data record) in the file
// ascat_next_node():     return node data for the next node in the current MDR
// ascat_get_sat():       return satellite id
// ascat_get_fmt():       return format id
// ascat_get_tot_mdr():   return number of MDRs in the file
// ascat_get_tot_node():  return number of nodes in the file
// ascat_get_node_beam(): return number of nodes per beam
// ascat_get_node_mdr():  return number of nodes per MDR
// ascat_clear_node():    initialize all of the node members to zero
// ascat_close_file():    close the file and do any cleanup needed
//
// Some private functions exist behind the scenes.
// Also some structs are used to contain all of the relevant data together.
//
// The basic structure of an L1B file is as follows:
// There is a Main Product Header Record (MPHR) that contains information about the rest of the file
// There are other headers that we mostly ignore
// Then there is a bunch of Measurement Data Records (MDR)
// For L1B format version 11 and below, each MDR contains the measurements of a single sequence of the six beams
// For L1B format version 12 and above, each MDR contains the measurements for only a single beam, so there are six times as many MDRs now
// This distinction is abstracted behind the ascat_next_node() function, just keep calling it until it returns false (0)
//
/********************************************************************/

// For the Fortran interface, add these lines to the Fortran source:
// (TODO: This is a bit outdated, also use proper interfaces to type-check function arguments)
// logical :: ascat_check_file, ascat_open_file
// integer :: ascat_get_tot_mdr, ascat_get_tot_node, ascat_next_mdr, ascat_next_node, ascat_get_node_beam, ascat_get_node_mdr
// external :: ascat_check_file, ascat_open_file, ascat_close_file, ascat_get_tot_mdr, ascat_get_tot_node, ascat_get_fmt
// external :: ascat_next_mdr, ascat_next_node, ascat_get_node_beam, ascat_get_node_mdr
// type ascat_node
//    integer*4 :: sat, fmt
//    logical   :: asc
//    integer*4 :: swath, index
//    real*8 :: lon, lat, tm, track, sigma0, inc, azi
//    real*8 :: beam_prev_lon, beam_prev_lat, beam_next_lon, beam_next_lat
//    real*8 :: track_prev_lon, track_prev_lat, track_next_lon, track_next_lat
// end type ascat_node

/********************************************************************/
// Constants

#define METOP1 1
#define METOP2 2
#define METOP3 3
#define METOPN 0

#define SZO 0
#define SZR 1
#define SZF 2
#define SZO_ID    2
#define SZR_ID    1
#define SZF_ID    3

/********************************************************************/
// Structures

// L1B file-level data
typedef struct {
 FILE *fp;               // file pointer
 size_t nodes_per_beam;  // number of nodes per beam
 size_t num_mdrs;        // number of MDRs in the file
 size_t cur_mdr;         // the current MDR number we're on (0..num_mdrs-1)
 unsigned char *mdrs;    // allocated buffer containing all the MDRs (the L1B file minus the headers before the first MDR)
 unsigned char *mdr;     // pointer to current MDR in mdrs buffer
 size_t mdr_size;        // size of each MDR (in bytes)
 unsigned int sat;       // host platform (note that MetOp-A is MetOp-2 and MetOp-B is MetOp-1)
 unsigned int fmt;       // product format (SZO, SZR, SZF)
 unsigned int fmt_id;    // record subclass ID (SZO: 2, SZR: 1, SZF: 3)
 unsigned int pver;      // processor version
 unsigned int fver;      // L1B format version (only 11 and 12 are presently known)
} ascat_file_t;

// L1B node-level data
//Note that this is meant for SZF data and not for SZO/SZR, but it could be expanded for those if needed
typedef struct {
 unsigned int sat;      // host platform
 unsigned int fmt;      // product format
 unsigned int asc;      // ascending flag: 1 for ascending, 0 for descending
 unsigned int swath;    // swath (1,2) for SZO/SZR (left/right) and selects beam (1-6) for SZF
 unsigned int index;    // the node index (eg, 1-256 for old format SZF, 1-192 for new format SZF)
 unsigned int ab_index; // the along-beam node index (eg, 1-256 for old format SZF, 33-224 for new format SZF mid beams, 39-230 for new format SZF side beams)
 double lon,lat;        // lat (-90 to 90) and lon (0 to 360) in degrees
 double tm;             // epoch time in days since 1 Jan 2000
 double track;          // azimuth angle bearing (0-360) relative to nadir track velocity
 double sigma0;         // sigma0 value (dB) at the node
 double inc;            // incidence angle (deg) at the node
 double azi;            // azimuth angle (deg, -180 to 180, where - is west and + is east) going from the measurement to nadir
 // The following values are populated if possible
 // They are the location (in lat/lon) of the previous and next measurement either
 // 1) neighboring along-beam measurements, same along-track beam
 // 2) neighboring along-track beams, same along-beam index
 double beam_prev_lon, beam_prev_lat; // The location of the previous node in the beam (if available)
 double beam_next_lon, beam_next_lat; // The location of the next node in the beam (if available)
 double track_prev_lon, track_prev_lat; // The location of the previous beam, same along-beam index (if available)
 double track_next_lon, track_next_lat; // The location of the next beam, same along-beam index (if available)
} ascat_node_t;

// ASCAT orbit information (extracted from MPHR)
typedef struct {
  char time_start[16], time_stop[16], state_time[19];
  double semi, eccen, inclination, perigee, ascension, mean_anomaly;
  double x_position, y_position, z_position;
  double x_vel, y_vel, z_vel;
  double startlat,startlon,stoplat,stoplon;
  int orbit_start, orbit_end;
} ascat_orbit_t;


/********************************************************************/
// Function prototypes
int ascat_check_file(const char *fname);
int ascat_open_file(const char *fname, ascat_file_t *a, ascat_orbit_t *orb);
void ascat_close_file(ascat_file_t *a);
size_t ascat_get_tot_mdr(const ascat_file_t *a);
size_t ascat_get_tot_node(const ascat_file_t *a);
size_t ascat_get_node_beam(const ascat_file_t *a);
size_t ascat_get_node_mdr(const ascat_file_t *a);
char *ascat_get_fmt(const ascat_file_t *a);
unsigned int ascat_get_sat(const ascat_file_t *a);
int ascat_next_mdr(ascat_file_t * const a, ascat_node_t * const node);
int ascat_next_node(const ascat_file_t * const a, ascat_node_t * const node);
void ascat_clear_node(ascat_node_t *node);

// Typical usage
//
// For a given L1B file, it's best to run ascat_check_file() on it to verify
// its correctness. Then ascat_open_file() is run to open the file. When
// finished, ascat_close_file() closes it.
// The ascat_get_*() functions are used to query some of the values in the
// ascat_file_t struct.
// Each L1B file is divided into several MDRs (measurement data records). To
// read all of the data in an L1B file, iterate through each MDR using
// ascat_next_mdr(), and for each MDR iterate through all of the measurements
// (nodes) using ascat_next_node().
// ascat_clear_node() is a utility function to clear out the variables in the
// ascat_node_t struct.
/********************************************************************/
// vim:et:tabstop=2:softtabstop=2:shiftwidth=2:
